from collections import defaultdict
from contextlib import contextmanager
from enum import Enum
import inspect
import json
import os
from .recorders import (
    ApiClientRecorder,
    Checkpoint,
    ComparisonResult,
    DiskRecorder,
    FileRecorder,
    MemoryRecorder,
    EvalRecorder,
    get_input_hash,
)
from .comparators import model_graded_comparator, output_feedback_from_scores
from .classes import (
    FeedbackResult,
    OutputComparator,
    OutputScore,
    OutputScorer,
    OutputFeedback,
)

import random
import threading
from typing import (
    Callable,
    Iterable,
    Iterator,
    Protocol,
    TypeVar,
    Optional,
    Any,
    runtime_checkable,
)
from functools import wraps
from dataclasses import dataclass
import dataclasses
import datetime
from .utils import get_user_input
from .classes import FeedbackResult
from rich.console import Console
from rich.table import Table

GREEN = "\x1b[32m"
ORANGE = "\x1b[33m"
RED = "\x1b[31m"
BOLD = "\x1b[1m"
UNDERLINE = "\x1b[4m"
ITALIC = "\x1b[3m"
HEADING_BG = "\x1b[103m"

END_CLR = "\x1b[0m"


@runtime_checkable
class JsonSerializable(Protocol):
    def to_json(self) -> str:
        ...

    @staticmethod
    def from_json(json_str: str) -> "JsonSerializable":
        ...


T = TypeVar("T", int, float, str, bool, bytes, dict, list, None, JsonSerializable)


@dataclass
class EvaluatorConfig:
    sample_rate: float = 1.0
    dir_path: str = "./eyeball_data"
    api_key: Optional[str] = None
    api_url: Optional[str] = None

    @staticmethod
    def _merge(original_config: "EvaluatorConfig", **kwargs) -> "EvaluatorConfig":
        """Kwargs should be a subset of the fields of the original config."""
        if len(kwargs) == 0:
            return original_config

        new_config = EvaluatorConfig()

        for field in dataclasses.fields(original_config):
            if (field_val := kwargs.get(field.name)) is not None:
                setattr(new_config, field.name, field_val)
            else:
                setattr(new_config, field.name, getattr(original_config, field.name))

        return new_config


class EvaluatorMode(Enum):
    RECORD = "record"
    COMPARE_CHECKPOINTS = "compare_checkpoints"
    RERUN_EXAMPLES = "rerun_examples"
    RATE_EXAMPLES = "rate_examples"


class Evaluator:
    def __init__(self, **config_kwargs) -> None:
        self.config = EvaluatorConfig()
        self.mode: EvaluatorMode = EvaluatorMode.RECORD
        self.current_recorder_state = threading.local()
        self.set_config(**config_kwargs)

    def _get_config(self, **override_config_kwargs) -> EvaluatorConfig:
        return EvaluatorConfig._merge(self.config, **override_config_kwargs)

    def set_config(self, **config_kwargs) -> None:
        self.config = EvaluatorConfig._merge(self.config, **config_kwargs)
        self.data_dir = os.path.join(self.config.dir_path, "eyeball_data")
        if self.config.api_key is not None:
            self.recorder: EvalRecorder = ApiClientRecorder(
                api_key=self.config.api_key, api_url=self.config.api_url
            )
        else:
            self.recorder = FileRecorder(self.data_dir)

    @contextmanager
    def start_recording_session(
        self,
        task_name: str,
        checkpoint_id: Optional[str] = None,
        checkpoint_id_to_rerun: Optional[str] = None,
    ) -> Iterator[None]:
        if hasattr(self.current_recorder_state, "recorder_checkpoint_id"):
            # This should not happen but if it does, we should not overwrite the previous checkpoint id
            # Instead we should set it to None so there is no confusion
            # If the user calls get_eval_params this will raise an error
            self.current_recorder_state.task_name = None
            self.current_recorder_state.recorder_checkpoint_id = None
            self.current_recorder_state.checkpoint_id_to_rerun = None
            yield None
            return

        self.current_recorder_state.task_name = task_name
        if checkpoint_id is not None:
            self.current_recorder_state.recorder_checkpoint_id = checkpoint_id
        else:
            self.current_recorder_state.recorder_checkpoint_id = (
                datetime.datetime.utcnow().isoformat()
            )
        self.current_recorder_state.checkpoint_id_to_rerun = checkpoint_id_to_rerun
        yield None

        del self.current_recorder_state.task_name
        del self.current_recorder_state.recorder_checkpoint_id

    def _should_record(self, **config_kwargs) -> bool:
        if self.mode == EvaluatorMode.COMPARE_CHECKPOINTS:
            return False

        if self.mode == EvaluatorMode.RERUN_EXAMPLES:
            return True

        config = self._get_config(**config_kwargs)
        if config.sample_rate == 0:
            return False
        return config.sample_rate == 1.0 or random.random() < config.sample_rate

    def record_input(
        self,
        variable_name: str,
        value: T,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
        **config_kwargs,
    ) -> None:
        if self._should_record(**config_kwargs):
            task_name, checkpoint_id = self._get_recorder_state(
                task_name, checkpoint_id
            )

            serialized_value = self._serialize(value)
            self.recorder.record_input_variable(
                task_name=task_name,
                checkpoint_id=checkpoint_id,
                variable_name=variable_name,
                value=serialized_value,
            )

    def record_output(
        self,
        value: Any,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
        **config_kwargs,
    ) -> None:
        if self._should_record(**config_kwargs):
            task_name, checkpoint_id = self._get_recorder_state(
                task_name, checkpoint_id
            )
            self.recorder.record_output(
                task_name=task_name,
                checkpoint_id=checkpoint_id,
                output=self._serialize(value),
            )

    def _get_last_checkpoint(
        self,
        task_name: str,
        input_hash: str,
        current_checkpoint_id: str,
    ) -> Optional[Checkpoint]:
        checkpoints = self.recorder.get_latest_checkpoints(
            task_name, input_hash, num_checkpoints=2
        )
        if len(checkpoints) == 0:
            return None

        if checkpoints[0].checkpoint_id == current_checkpoint_id:
            if len(checkpoints) == 1:
                return None
            return checkpoints[1]

        else:
            return checkpoints[0]

    def _get_recorder_state(
        self,
        task_name: Optional[str] = None,
        recorder_checkpoint_id: Optional[str] = None,
    ) -> tuple[str, str]:
        if task_name is None and hasattr(self.current_recorder_state, "task_name"):
            task_name = self.current_recorder_state.task_name
        if task_name is None:
            raise ValueError(
                "Must provide a task name or start a recording session to record a variable"
            )

        if recorder_checkpoint_id is None and hasattr(
            self.current_recorder_state, "recorder_checkpoint_id"
        ):
            recorder_checkpoint_id = self.current_recorder_state.recorder_checkpoint_id
        if recorder_checkpoint_id is None:
            raise ValueError(
                "Must provide a checkpoint id or start a recording session to record a variable"
            )

        return task_name, recorder_checkpoint_id

    def get_eval_params(
        self,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
    ) -> Optional[dict[str, Any]]:
        task_name, checkpoint_id = self._get_recorder_state(task_name, checkpoint_id)

        checkpoint = self.recorder.get_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        if checkpoint is None:
            return None
        return checkpoint.eval_params

    def get_eval_param(
        self,
        param_name: str,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
    ) -> Optional[Any]:
        params = self.get_eval_params(task_name=task_name, checkpoint_id=checkpoint_id)
        if params is None:
            return None
        return params.get(param_name)

    def _serialize(self, value: T) -> str:
        if isinstance(value, JsonSerializable):
            return value.to_json()
        else:
            return json.dumps(value)

    def record_task(
        self,
        args_to_record: list[str],
        task_name: Optional[str] = None,
        checkpoint_id_arg_name: Optional[str] = None,
        eval_params: Optional[dict[str, Any]] = None,
        request_user_feedback_probability: float = 0.0,
        **config_kwargs,
    ) -> Callable[..., Callable]:
        def _decorator(fn: Callable[..., T]) -> Callable[..., T]:
            @wraps(fn)
            def _wrapper(*args, **kwargs) -> Any:
                if not self._should_record(**config_kwargs):
                    return fn(*args, **kwargs)

                (
                    fn_arg_names,
                    _,
                    _,
                    _,
                    _,
                    _,
                    _,
                ) = inspect.getfullargspec(fn)

                # Set task name
                if self.mode == EvaluatorMode.RERUN_EXAMPLES:
                    local_task_name = self.current_recorder_state.task_name
                else:
                    local_task_name = task_name or fn.__name__

                # Set checkpoint id
                if self.mode == EvaluatorMode.RERUN_EXAMPLES:
                    recorder_checkpoint_id = (
                        self.current_recorder_state.recorder_checkpoint_id
                    )
                elif checkpoint_id_arg_name is not None:
                    if checkpoint_id_arg_name in kwargs:
                        recorder_checkpoint_id = kwargs[checkpoint_id_arg_name]
                    else:
                        try:
                            checkpoint_id_arg_index = fn_arg_names.index(
                                checkpoint_id_arg_name
                            )
                            recorder_checkpoint_id = args[checkpoint_id_arg_index]
                        except ValueError:
                            recorder_checkpoint_id = (
                                datetime.datetime.utcnow().isoformat()
                            )
                else:
                    recorder_checkpoint_id = datetime.datetime.utcnow().isoformat()

                # Set params
                if self.mode == EvaluatorMode.RERUN_EXAMPLES:
                    params = self.get_eval_params(
                        task_name=local_task_name,
                        checkpoint_id=recorder_checkpoint_id,
                    )
                    # TODO: maybe we should not override with default params in this mode. Can lead to some confusion
                    if (params is None or len(params) == 0) and eval_params is not None:
                        self.recorder.record_eval_params(
                            task_name=local_task_name,
                            checkpoint_id=recorder_checkpoint_id,
                            eval_params=eval_params,
                        )
                elif eval_params is not None:
                    self.recorder.record_eval_params(
                        task_name=local_task_name,
                        checkpoint_id=recorder_checkpoint_id,
                        eval_params=eval_params,
                    )

                for arg_name, arg_val in zip(fn_arg_names, args):
                    if arg_name in args_to_record:
                        self.record_input(
                            task_name=local_task_name,
                            checkpoint_id=recorder_checkpoint_id,
                            variable_name=arg_name,
                            value=arg_val,
                        )

                for kwarg_name, kwarg_val in kwargs.items():
                    if kwarg_name in args_to_record:
                        self.record_input(
                            task_name=local_task_name,
                            checkpoint_id=recorder_checkpoint_id,
                            variable_name=kwarg_name,
                            value=kwarg_val,
                        )

                if self.mode == EvaluatorMode.RERUN_EXAMPLES:
                    result = fn(*args, **kwargs)  # type: ignore
                else:
                    with self.start_recording_session(
                        task_name=local_task_name,
                        checkpoint_id=recorder_checkpoint_id,
                    ):
                        result = fn(*args, **kwargs)  # type: ignore

                self.record_output(
                    task_name=local_task_name,
                    checkpoint_id=recorder_checkpoint_id,
                    value=result,
                )
                if request_user_feedback_probability > 0:
                    if random.random() < request_user_feedback_probability:
                        print("Requesting user feedback")
                        self.request_user_feedback(
                            output=str(result),
                            task_name=local_task_name,
                            checkpoint_id=recorder_checkpoint_id,
                        )
                return result

            return _wrapper

        return _decorator

    def request_user_feedback(
        self,
        output: str,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
        print_inputs: bool = False,
    ) -> Optional[OutputFeedback]:
        task_name, checkpoint_id = self._get_recorder_state(task_name, checkpoint_id)
        checkpoint = self.recorder.get_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        if checkpoint is None:
            return None

        if print_inputs:
            print(checkpoint.get_input_var_str())

        feedback_result = FeedbackResult[
            get_user_input(
                f"What do you think of the result?\n{output}\n",
                choices=[f.name for f in FeedbackResult],
            )
        ]
        details = get_user_input(
            "Any details about your feedback (optional)\n", choices=None
        )

        output_feedback = OutputFeedback(result=feedback_result, message=details)

        self.recorder.record_output_feedback(
            task_name=task_name, checkpoint_id=checkpoint_id, feedback=output_feedback
        )
        return output_feedback

    def rerun_recorded_examples(
        self,
        *eval_params_list: dict[str, Any],
        task_name: Optional[str] = None,
        input_hashes: Optional[list[str]] = None,
    ) -> Iterable[dict[str, Any]]:
        if task_name is None:
            task_names = self.recorder.get_task_names()
            if len(task_names) == 0:
                raise ValueError("No task names found")
            elif len(task_names) == 1:
                task_name = task_names[0]
            else:
                raise ValueError(
                    f"Must provide a task_name as you have multiple tasks recorded -- Found: {task_names}"
                )

        if input_hashes is None or len(input_hashes) == 0:
            input_hashes = self.recorder.get_input_hashes(task_name=task_name)

        try:
            self.mode = EvaluatorMode.RERUN_EXAMPLES
            rerun_metadata = {
                "id": datetime.datetime.utcnow().isoformat(),
            }

            print(f"Will rerun {len(input_hashes)} inputs for task:`{task_name}`")

            for idx, input_hash in enumerate(input_hashes):
                recorder_checkpoint_id = datetime.datetime.utcnow().isoformat()
                last_checkpoint = self._get_last_checkpoint(
                    task_name=task_name,
                    input_hash=input_hash,
                    current_checkpoint_id=recorder_checkpoint_id,
                )
                if last_checkpoint is None:
                    continue

                self.checkpoint_id_to_rerun = last_checkpoint.checkpoint_id
                checkpoint_to_rerun = last_checkpoint
                if checkpoint_to_rerun is None:
                    continue

                input_vars = {
                    k: json.loads(v)
                    for k, v in checkpoint_to_rerun.get_input_variables().items()
                }
                print(
                    f"\nRerunning input #{idx}:\n{checkpoint_to_rerun.get_input_var_str()}"
                )

                if len(eval_params_list) > 0:
                    for eval_params in eval_params_list:
                        with self.start_recording_session(
                            task_name=task_name,
                            checkpoint_id=recorder_checkpoint_id,
                            checkpoint_id_to_rerun=checkpoint_to_rerun.checkpoint_id,
                        ):
                            self.recorder.record_eval_params(
                                task_name=task_name,
                                checkpoint_id=recorder_checkpoint_id,
                                eval_params=eval_params,
                                rerun_metadata=rerun_metadata,
                            )
                            print(f"Using eval params: {eval_params}")
                            yield input_vars
                else:
                    with self.start_recording_session(
                        task_name=task_name,
                        checkpoint_id=recorder_checkpoint_id,
                    ):
                        self.recorder.record_eval_params(
                            task_name=task_name,
                            checkpoint_id=recorder_checkpoint_id,
                            eval_params={},
                            rerun_metadata=rerun_metadata,
                        )
                        yield input_vars
        finally:
            self.mode = EvaluatorMode.RECORD

    def rate_recorded_examples(
        self,
        *input_hashes: str,
        task_name: Optional[str] = None,
    ) -> None:
        if task_name is None:
            task_names = self.recorder.get_task_names()
            if len(task_names) == 0:
                raise ValueError("No task names found")
            elif len(task_names) == 1:
                task_name = task_names[0]
            else:
                raise ValueError(
                    f"Must provide a task_name as you have multiple tasks recorded -- Found: {task_names}"
                )

        input_hashes_lst: list[str] = list(input_hashes)

        if len(input_hashes_lst) == 0:
            input_hashes_lst = self.recorder.get_input_hashes(task_name=task_name)

        try:
            self.mode = EvaluatorMode.RATE_EXAMPLES
            for input_hash in input_hashes_lst:
                already_seen_outputs_to_feedback: dict[str:OutputFeedback] = {}
                checkpoints = self.recorder.get_latest_checkpoints(
                    task_name, input_hash, num_checkpoints=4
                )

                if len(checkpoints) == 0:
                    continue

                print(f"For the inputs:\n{checkpoints[0].get_input_var_str()}")
                for checkpoint in checkpoints:
                    if checkpoint.output_feedback is None:
                        if feedback := already_seen_outputs_to_feedback.get(
                            checkpoint.output
                        ):
                            self.recorder.record_output_feedback(
                                task_name=task_name,
                                checkpoint_id=checkpoint.checkpoint_id,
                                feedback=feedback,
                            )
                        else:
                            feedback = self.request_user_feedback(
                                checkpoint.output or "",
                                task_name=task_name,
                                checkpoint_id=checkpoint.checkpoint_id,
                            )
                            if feedback is not None:
                                already_seen_outputs_to_feedback[
                                    checkpoint.output
                                ] = feedback
                    elif new_feedback := already_seen_outputs_to_feedback.get(
                        checkpoint.output
                    ):
                        if new_feedback.result != checkpoint.output_feedback.result:
                            print(
                                f"For output: {checkpoint.output}\nOld feedback {checkpoint.output_feedback} is different from new feedback {new_feedback}"
                            )
                            print(f"Updating feedback to {new_feedback}")
                            self.recorder.record_output_feedback(
                                task_name=task_name,
                                checkpoint_id=checkpoint.checkpoint_id,
                                feedback=new_feedback,
                            )
                    else:
                        print(f"Already has feedback: {checkpoint.output_feedback}")
        finally:
            self.mode = EvaluatorMode.RECORD

    def record_human_feedback(
        self,
        feedback: OutputFeedback,
        details: Optional[str] = None,
        task_name: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
    ):
        task_name, checkpoint_id = self._get_recorder_state(task_name, checkpoint_id)
        pass

    def delete_checkpoints_for_input_vars(
        self,
        input_vars: dict[str, str],
        task_name: Optional[str] = None,
    ) -> None:
        if task_name is None:
            task_names = self.recorder.get_task_names()
            if len(task_names) == 0:
                raise ValueError("No task names found")
            elif len(task_names) == 1:
                task_name = task_names[0]
            else:
                raise ValueError(
                    f"Must provide a task_name as you have multiple tasks recorded -- Found: {task_names}"
                )
        input_hash = get_input_hash(
            {k: self._serialize(v) for k, v in input_vars.items()}
        )
        self.recorder.delete_checkpoints_for_input_hash(task_name, input_hash)

    def compute_latest_comparison_results(self, task_name: str) -> None:
        input_names: set[str] = set()
        rows: list[dict[str, str]] = []
        comparison_column_names = [
            "latest_checkpoint",
            "previous_checkpoint",
            "the_checkpoint_before",
        ]

        for input_hash in self.recorder.get_input_hashes(task_name):
            checkpoints = self.recorder.get_latest_checkpoints(
                task_name, input_hash, num_checkpoints=1
            )
            if len(checkpoints) == 0:
                continue

            checkpoint = checkpoints[0]
            if checkpoint is None:
                continue

            row_data = {}
            input_names.update(checkpoint.input_variables.keys())
            for input_name, input_value in checkpoint.input_variables.items():
                row_data[input_name] = input_value

            comparison_results = self.recorder.get_comparison_results_for_input_hash(
                task_name=task_name,
                input_hash=input_hash,
                num_results=len(comparison_column_names),
            )
            sorted_comparison_results = sorted(
                comparison_results,
                key=lambda x: x.newer_checkpoint_id,
                reverse=True,
            )
            for idx, comparison_result in enumerate(sorted_comparison_results):
                msg = str(comparison_result.output_feedback.result)
                new_checkpoint = self.recorder.get_checkpoint(
                    task_name=task_name,
                    checkpoint_id=comparison_result.newer_checkpoint_id,
                )
                if new_checkpoint is not None:
                    if new_checkpoint.output_score is not None:
                        msg += f" (score: {new_checkpoint.output_score})"
                    eval_params_str = ", ".join(
                        f"{k}={v}" for k, v in new_checkpoint.eval_params.items()
                    )
                    if eval_params_str:
                        msg += f" ({eval_params_str})"
                row_data[comparison_column_names[idx]] = msg

            rows.append(row_data)

        md_data = []
        table = Table(title="Comparison Results")

        if len(rows) > 0:
            column_names = sorted(list(input_names)) + comparison_column_names
            for column_name in column_names:
                table.add_column(column_name, justify="left")

            md_data.append("| " + " | ".join(column_names) + " |")
            md_data.append("| " + " | ".join(["---"] * len(column_names)) + " |")

            for row in rows:
                row_tuple = tuple(
                    row.get(column_name, "") for column_name in column_names
                )
                table.add_row(*row_tuple)
                md_data.append("| " + " | ".join(row_tuple) + " |")

            task_data_dir = os.path.join(self.data_dir, task_name)
            if not os.path.exists(task_data_dir):
                os.makedirs(task_data_dir)

            with open(
                os.path.join(task_data_dir, "benchmark.md"),
                "w+",
            ) as f:
                f.write("\n".join(md_data))

            console = Console()
            console.print(table)

    def compare_recorded_checkpoints(
        self,
        task_name: Optional[str] = None,
        num_input_hashes: int = 5,
        num_checkpoints_per_input_hash: int = 3,
        task_objective: Optional[str] = None,
        output_comparator: Optional[OutputComparator] = None,
        output_scorer: Optional[OutputScorer] = None,
        use_cached_scores: bool = True,
        use_cached_comparisons: bool = True,
    ) -> None:
        if task_name is None:
            task_names = self.recorder.get_task_names()
            if len(task_names) == 0:
                raise ValueError("No task names found")
            elif len(task_names) == 1:
                task_name = task_names[0]
            else:
                raise ValueError(
                    f"Must provide a task_name as you have multiple tasks recorded -- Found: {task_names}"
                )

        if output_comparator is None and output_scorer is None:
            output_comparator = model_graded_comparator
            assert (
                task_objective is not None
            ), "Must provide an objective to compare or an output comparator"

        input_hashes_list = self.recorder.get_input_hashes(task_name)
        random.shuffle(input_hashes_list)
        input_hashes_list = input_hashes_list[:num_input_hashes]

        print(f"Comparing {len(input_hashes_list)} inputs for task:`{task_name}`")

        try:
            self.mode = EvaluatorMode.COMPARE_CHECKPOINTS

            params_to_succesful_examples: dict[str, int] = defaultdict(int)
            rerun_to_succesful_examples: dict[str, int] = defaultdict(int)

            num_comparisons = 0
            for idx, input_hash in enumerate(input_hashes_list):
                checkpoints_list = self.recorder.get_latest_checkpoints(
                    task_name,
                    input_hash,
                    num_checkpoints=num_checkpoints_per_input_hash,
                )

                checkpoints: dict[str, Checkpoint] = {
                    c.checkpoint_id: c
                    for c in checkpoints_list
                    if c.output is not None
                    and c.output != "null"
                    and c.output != "None"
                }
                checkpoint_ids = sorted(checkpoints.keys(), reverse=True)

                if len(checkpoint_ids) == 0:
                    continue

                scores: dict[str, OutputScore] = {}
                if output_scorer is not None:
                    print(f"\nInput #{idx} - Scoring {len(checkpoint_ids)} checkpoints")
                    for checkpoint_id in checkpoint_ids:
                        checkpoint_to_score = checkpoints[checkpoint_id]
                        if (
                            use_cached_scores
                            and checkpoint_to_score.output_score is not None
                        ):
                            print(
                                f"Using cached score for {checkpoint_id}: {checkpoint_to_score.output_score}"
                            )
                            scores[checkpoint_id] = checkpoint_to_score.output_score
                            continue

                        assert checkpoint_to_score.output is not None
                        score = output_scorer(
                            task_objective or "",
                            checkpoint_to_score.get_input_variables(),
                            checkpoint_to_score.output,
                        )
                        self.recorder.record_output_score(
                            task_name=task_name,
                            checkpoint_id=checkpoint_id,
                            score=score,
                        )
                        scores[checkpoint_id] = score
                        print(f"Scored {checkpoint_id}: {score}")

                if len(checkpoint_ids) < 2:
                    continue

                # Newer checkpoints are at the start of the list so for every comparison
                # we compare the newer_checkpoint_id with the older_checkpoint_id
                to_compare = [
                    (checkpoint_ids[i], checkpoint_ids[i + 1])
                    for i in range(len(checkpoint_ids) - 1)
                ]

                print(f"\nInput #{idx} - Running {len(to_compare)} comparison(s)")

                for newer_checkpoint_id, older_checkpoint_id in to_compare:
                    newer_checkpoint = self.recorder.get_checkpoint(
                        task_name=task_name,
                        checkpoint_id=newer_checkpoint_id,
                    )
                    older_checkpoint = self.recorder.get_checkpoint(
                        task_name=task_name,
                        checkpoint_id=older_checkpoint_id,
                    )

                    if newer_checkpoint is None or older_checkpoint is None:
                        print(
                            f"Could not find checkpoint {newer_checkpoint_id} or {older_checkpoint_id}"
                        )
                        continue

                    newer_checkpoint_output = newer_checkpoint.output
                    older_checkpoint_output = older_checkpoint.output

                    assert newer_checkpoint_output is not None
                    assert older_checkpoint_output is not None

                    should_record_comparison = True
                    if use_cached_comparisons and (
                        comparator_result := self.recorder.get_comparison_result(
                            task_name=task_name,
                            older_checkpoint_id=older_checkpoint_id,
                            newer_checkpoint_id=newer_checkpoint_id,
                        )
                    ):
                        print(f"Using cached comparison result for {input_hash}")
                        should_record_comparison = False
                        output_comparison_feedback = comparator_result.output_feedback
                    elif output_comparator is not None:
                        output_comparison_feedback = output_comparator(
                            objective=task_objective or "",
                            input_variables=newer_checkpoint.get_input_variables(),
                            older_checkpoint_output=older_checkpoint_output,
                            newer_checkpoint_output=newer_checkpoint_output,
                        )
                    elif output_scorer is not None:
                        output_comparison_feedback = output_feedback_from_scores(
                            older_score=scores[older_checkpoint_id],
                            newer_score=scores[newer_checkpoint_id],
                        )
                    else:
                        raise ValueError(
                            "Should not happen. We need an output comparator or output scorer"
                        )

                    new_unique_params = []
                    old_unqiue_params = []
                    for key in (
                        newer_checkpoint.eval_params.keys()
                        | older_checkpoint.eval_params.keys()
                    ):
                        param_val_a = newer_checkpoint.eval_params.get(key)
                        param_val_b = older_checkpoint.eval_params.get(key)
                        if param_val_a != param_val_b:
                            if param_val_a is not None:
                                new_unique_params.append(f"{key}={param_val_a}")
                            elif param_val_b is not None:
                                old_unqiue_params.append(f"{key}={param_val_b}")

                    new_unique_params_str = (
                        ("(" + ", ".join(new_unique_params) + ")")
                        if len(new_unique_params) > 0
                        else ""
                    )
                    old_unique_params_str = (
                        ("(" + ", ".join(old_unqiue_params) + ")")
                        if len(old_unqiue_params) > 0
                        else ""
                    )

                    new_params_str = ",".join(
                        f"{k}={v}" for k, v in newer_checkpoint.eval_params.items()
                    )
                    old_params_str = ",".join(
                        f"{k}={v}" for k, v in older_checkpoint.eval_params.items()
                    )

                    rerun_id_new = (
                        newer_checkpoint.rerun_metadata.get("id")
                        if newer_checkpoint.rerun_metadata is not None
                        else None
                    )
                    rerun_id_old = (
                        older_checkpoint.rerun_metadata.get("id")
                        if older_checkpoint.rerun_metadata is not None
                        else None
                    )

                    num_comparisons += 1

                    if output_comparison_feedback.result == FeedbackResult.NEUTRAL:
                        print(
                            f"{ORANGE}[neutral] task output is the same between checkpoints {older_checkpoint_id} {old_unique_params_str} & {newer_checkpoint_id} {new_unique_params_str} {END_CLR}"
                        )
                        params_to_succesful_examples[new_params_str] += 1
                        params_to_succesful_examples[old_params_str] += 1
                        if rerun_id_new is not None:
                            rerun_to_succesful_examples[rerun_id_new] += 1
                        if rerun_id_old is not None:
                            rerun_to_succesful_examples[rerun_id_old] += 1
                    elif output_comparison_feedback.result == FeedbackResult.NEGATIVE:
                        print(
                            f"{RED}[regression] task output was better in the older checkpoint {older_checkpoint_id} {old_unique_params_str} than {newer_checkpoint_id} {new_unique_params_str} {END_CLR}"
                        )
                        params_to_succesful_examples[old_params_str] += 1
                        if rerun_id_old is not None:
                            rerun_to_succesful_examples[rerun_id_old] += 1
                    else:
                        print(
                            f"{GREEN}[improvement] task output improved from checkpoint {older_checkpoint_id} {old_unique_params_str} to {newer_checkpoint_id} {new_unique_params_str}{END_CLR}"
                        )
                        params_to_succesful_examples[new_params_str] += 1
                        if rerun_id_new is not None:
                            rerun_to_succesful_examples[rerun_id_new] += 1
                    if should_record_comparison:
                        self.recorder.record_comparison_result(
                            task_name=task_name,
                            input_hash=input_hash,
                            result=ComparisonResult(
                                older_checkpoint_id=older_checkpoint_id,
                                newer_checkpoint_id=newer_checkpoint_id,
                                output_feedback=output_comparison_feedback,
                            ),
                        )

            print("\nSummary:")
            print("---------")
            if len(rerun_to_succesful_examples) > 0:
                print("Your most sucessful re-runs:")
                for rerun_id, num_successes in sorted(
                    rerun_to_succesful_examples.items(),
                    key=lambda x: x[1],
                    reverse=True,
                ):
                    print(
                        f"{rerun_id}: {num_successes}/{len(input_hashes_list)} successes"
                    )

            print("\nYour most sucessful params:")
            for params, num_successes in sorted(
                params_to_succesful_examples.items(), key=lambda x: x[1], reverse=True
            ):
                if params == "":
                    params = "default"
                print(f"{params}: {num_successes}/{num_comparisons} successes")

            self.compute_latest_comparison_results(task_name)

        finally:
            self.mode = EvaluatorMode.RECORD


_default_evaluator = Evaluator()


set_config = _default_evaluator.set_config
record_task = _default_evaluator.record_task
record_output = _default_evaluator.record_output
record_input = _default_evaluator.record_input
get_eval_param = _default_evaluator.get_eval_param
get_eval_params = _default_evaluator.get_eval_params
rerun_recorded_examples = _default_evaluator.rerun_recorded_examples
rate_recorded_examples = _default_evaluator.rate_recorded_examples
compare_recorded_checkpoints = _default_evaluator.compare_recorded_checkpoints
delete_checkpoints_for_input_vars = _default_evaluator.delete_checkpoints_for_input_vars



from .eval import (
    rerun_recorded_examples,
    set_config,
    record_task,
    record_output,
    record_input,
    get_eval_param,
    rate_recorded_examples,
    compare_recorded_checkpoints,
    delete_checkpoints_for_input_vars,
    Evaluator,
)

from .classes import (
    FeedbackResult,
    OutputFeedback,
    OutputComparator,
    OutputScorer,
    OutputScore,
)

__all__ = [
    "set_config",
    "record_task",
    "record_output",
    "record_input",
    "get_eval_param",
    "rerun_recorded_examples",
    "rate_recorded_examples",
    "compare_recorded_checkpoints",
    "delete_checkpoints_for_input_vars",
    "Evaluator",
    "FeedbackResult",
    "OutputFeedback",
    "OutputComparator",
    "OutputScorer",
    "OutputScore",
]



from typing import Any, Optional, Sequence
from collections import OrderedDict


def get_user_input(text: str, choices: Optional[Sequence[Any]]) -> str:
    from InquirerPy import inquirer
    import termios
    import sys

    termios.tcflush(sys.stdin, termios.TCIOFLUSH)

    if choices:
        return inquirer.select(text, choices=choices).execute()
    else:
        return inquirer.text(text).execute()


class LruCache(OrderedDict):
    def __init__(self, max_size: int = 100) -> None:
        super().__init__()
        self.max_size = max_size

    def __setitem__(self, key: Any, value: Any) -> None:
        if len(self) >= self.max_size:
            self.popitem(last=False)
        super().__setitem__(key, value)



from concurrent.futures import ThreadPoolExecutor
import hashlib
import json
import pkg_resources
import yaml
import os
import pickle
from typing import Any, Optional, Protocol
from dataclasses import dataclass
import dataclasses
import requests

from .utils import LruCache
from .classes import OutputFeedback, OutputScore
import logging

logger = logging.getLogger(__name__)


@dataclass
class ComparisonResult:
    older_checkpoint_id: str
    newer_checkpoint_id: str
    output_feedback: OutputFeedback

    def as_dict(self):
        return {
            "older_checkpoint_id": self.older_checkpoint_id,
            "newer_checkpoint_id": self.newer_checkpoint_id,
            "output_feedback": self.output_feedback.as_dict(),
        }


def get_input_hash(input_variables: dict[str, str]) -> str:
    sorted_input_vars = sorted(
        [
            (str(var_name), str(var_val))
            for var_name, var_val in input_variables.items()
        ],
        key=lambda x: x[0],
    )
    input_var_str = json.dumps(sorted_input_vars)
    return hashlib.sha256(input_var_str.encode("utf-8")).hexdigest()


@dataclass
class Checkpoint:
    checkpoint_id: str
    input_variables: dict[str, str] = dataclasses.field(default_factory=dict)
    eval_params: dict[str, Any] = dataclasses.field(default_factory=dict)
    output: Optional[str] = None
    output_feedback: Optional[OutputFeedback] = None
    output_score: Optional[OutputScore] = None
    rerun_metadata: dict[str, str] = dataclasses.field(default_factory=dict)

    def get_input_variables(self) -> dict[str, str]:
        return dict(self.input_variables)

    def get_input_hash(self) -> str:
        return get_input_hash(self.input_variables)

    def get_input_var_str(self) -> str:
        sorted_input_vars = sorted(
            [
                (str(var_name), str(var_val))
                for var_name, var_val in self.input_variables.items()
            ],
            key=lambda x: x[0],
        )
        msg = ""
        for var_name, var_val in sorted_input_vars:
            msg += f"{var_name}={var_val}\n"
        return msg

    def __str__(self) -> str:
        msg = f"Checkpoint: {self.checkpoint_id}\n"
        msg += self.get_input_var_str()
        return msg


class EvalRecorder(Protocol):
    def record_input_variable(
        self,
        task_name: str,
        checkpoint_id: str,
        variable_name: str,
        value: str,
    ) -> None:
        ...

    def record_output(
        self,
        task_name: str,
        checkpoint_id: str,
        output: str,
    ) -> None:
        ...

    def record_eval_params(
        self,
        task_name: str,
        checkpoint_id: str,
        eval_params: dict[str, Any],
        rerun_metadata: Optional[dict[str, Any]] = None,
    ) -> None:
        ...

    def record_output_feedback(
        self,
        task_name: str,
        checkpoint_id: str,
        feedback: OutputFeedback,
    ) -> None:
        ...

    def record_output_score(
        self,
        task_name: str,
        checkpoint_id: str,
        score: OutputScore,
    ) -> None:
        ...

    def get_checkpoint(
        self, task_name: str, checkpoint_id: str
    ) -> Optional[Checkpoint]:
        ...

    def get_latest_checkpoints(
        self, task_name: str, input_hash: str, num_checkpoints: int = 2
    ) -> list[Checkpoint]:
        ...

    def get_input_hashes(self, task_name: str) -> list[str]:
        ...

    def get_task_names(self) -> list[str]:
        ...

    def record_comparison_result(
        self,
        task_name: str,
        input_hash: str,
        result: ComparisonResult,
    ) -> None:
        ...

    def get_comparison_result(
        self,
        task_name: str,
        older_checkpoint_id: str,
        newer_checkpoint_id: str,
    ) -> Optional[ComparisonResult]:
        ...

    def get_comparison_results_for_input_hash(
        self, task_name: str, input_hash: str, num_results: int = 3
    ) -> list[ComparisonResult]:
        ...

    def delete_checkpoints_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
    ) -> None:
        ...


class ApiClientRecorder(EvalRecorder):
    def __init__(self, api_key: str, api_url: Optional[str] = None) -> None:
        self.api_key = api_key
        if api_url is None:
            self.url = "https://eyeball.tark.ai/"
        else:
            self.url = api_url
        self.checkpoint_dicts: LruCache = LruCache(max_size=100)
        self.pool = ThreadPoolExecutor(max_workers=1)

    def _get_headers(self) -> dict[str, str]:
        return {
            "Authorization": f"Bearer {self.api_key}",
        }

    def _record_checkpoint(
        self,
        task_name: str,
        checkpoint_id: str,
        prefixes: list[str],
        name: str,
        value: Any,
        flush: bool = False,
    ) -> dict[str, Any]:
        dict_key = f"{task_name},{checkpoint_id}"
        if dict_key not in self.checkpoint_dicts:
            checkpoint_dict: dict[str, Any] = {"checkpoint_id": checkpoint_id}
        else:
            checkpoint_dict = self.checkpoint_dicts[dict_key]

        current_dict = checkpoint_dict
        for prefix in prefixes:
            if prefix not in current_dict:
                current_dict[prefix] = {}
            current_dict = current_dict[prefix]
        current_dict[name] = value
        self.checkpoint_dicts[dict_key] = dict(checkpoint_dict)

        if flush:

            def _record():
                response = requests.post(
                    f"{self.url}/record_checkpoint",
                    json={
                        "task_name": task_name,
                        "checkpoint_id": checkpoint_id,
                        **checkpoint_dict,
                    },
                    headers=self._get_headers(),
                )
                if response.status_code != 200:
                    logger.error(
                        f"Failed to record checkpoint {checkpoint_id} -- {response.status_code}, {response.text}"
                    )
                else:
                    logger.debug(f"Recorded checkpoint {checkpoint_id}")

            self.pool.submit(_record)
        return checkpoint_dict

    def record_input_variable(
        self, task_name: str, checkpoint_id: str, variable_name: str, value: str
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=["input_variables"],
            name=variable_name,
            value=value,
        )

    def record_output(self, task_name: str, checkpoint_id: str, output: str) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="output",
            value=output,
            flush=True,
        )

    def record_eval_params(
        self,
        task_name: str,
        checkpoint_id: str,
        eval_params: dict[str, Any],
        rerun_metadata: dict[str, Any] | None = None,
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="eval_params",
            value=eval_params,
        )
        if rerun_metadata is not None:
            self._record_checkpoint(
                task_name=task_name,
                checkpoint_id=checkpoint_id,
                prefixes=[],
                name="rerun_metadata",
                value=rerun_metadata,
            )

    def record_comparison_result(
        self, task_name: str, input_hash: str, result: ComparisonResult
    ) -> None:
        requests.post(
            f"{self.url}/record_comparison_result",
            json={
                "task_name": task_name,
                "input_hash": input_hash,
                "result": result.as_dict(),
            },
            headers=self._get_headers(),
        )

    def record_output_score(
        self, task_name: str, checkpoint_id: str, score: OutputScore
    ) -> None:
        requests.post(
            f"{self.url}/record_output_score",
            json={
                "task_name": task_name,
                "checkpoint_id": checkpoint_id,
                "score": score.as_dict(),
            },
            headers=self._get_headers(),
        )

    def record_output_feedback(
        self, task_name: str, checkpoint_id: str, feedback: OutputFeedback
    ) -> None:
        requests.post(
            f"{self.url}/record_output_feedback",
            json={
                "task_name": task_name,
                "checkpoint_id": checkpoint_id,
                "feedback": feedback.as_dict(),
            },
            headers=self._get_headers(),
        )

    def delete_checkpoints_for_input_hash(
        self, task_name: str, input_hash: str
    ) -> None:
        ...

    def get_checkpoint(self, task_name: str, checkpoint_id: str) -> Checkpoint | None:
        dict_key = f"{task_name},{checkpoint_id}"
        if dict_key in self.checkpoint_dicts:
            return Checkpoint(**(self.checkpoint_dicts[dict_key]))

        response = requests.get(
            f"{self.url}/get_checkpoint",
            json={
                "task_name": task_name,
                "checkpoint_id": checkpoint_id,
            },
            headers=self._get_headers(),
        )
        if response.status_code != 200:
            logger.debug(
                f"Failed to get checkpoint - {checkpoint_id}: {response.status_code}"
            )
            return None
        return Checkpoint(**response.json())

    def get_comparison_result(
        self, task_name: str, older_checkpoint_id: str, newer_checkpoint_id: str
    ) -> ComparisonResult | None:
        ...

    def get_comparison_results_for_input_hash(
        self, task_name: str, input_hash: str, num_results: int = 3
    ) -> list[ComparisonResult]:
        return []

    def get_input_hashes(self, task_name: str) -> list[str]:
        response = requests.get(
            f"{self.url}/get_input_hashes",
            json={
                "task_name": task_name,
            },
            headers=self._get_headers(),
        )
        if response.status_code != 200:
            logger.debug(f"Failed to get input hashes: {response.status_code}")
            return []
        return response.json()["input_hashes"]

    def get_latest_checkpoints(
        self, task_name: str, input_hash: str, num_checkpoints: int = 2
    ) -> list[Checkpoint]:
        response = requests.get(
            f"{self.url}/get_latest_checkpoints",
            json={
                "task_name": task_name,
                "input_hash": input_hash,
                "num_checkpoints": num_checkpoints,
            },
            headers=self._get_headers(),
        )
        if response.status_code != 200:
            logger.debug(f"Failed to get latest checkpoints: {response.status_code}")
            return []
        return [Checkpoint(**data) for data in response.json()["checkpoints"]]

    def get_task_names(self) -> list[str]:
        response = requests.get(
            f"{self.url}/get_task_names",
            headers=self._get_headers(),
        )
        if response.status_code != 200:
            logger.debug(f"Failed to get task names: {response.status_code}")
            return []
        return response.json()["task_names"]


@dataclass
class Task:
    name: str
    checkpoints: dict[str, Checkpoint] = dataclasses.field(default_factory=dict)
    input_hashes: dict[str, set[str]] = dataclasses.field(default_factory=dict)
    comparison_results: dict[str, ComparisonResult] = dataclasses.field(
        default_factory=dict
    )
    input_hash_to_comparison_results: dict[str, set[str]] = dataclasses.field(
        default_factory=dict
    )


class MemoryRecorder(EvalRecorder):
    def __init__(self) -> None:
        self.tasks: dict[str, Task] = {}

    def _fetch_or_create_checkpoint(
        self, task_name: str, checkpoint_id: str
    ) -> Checkpoint:
        if task_name not in self.tasks:
            task = Task(name=task_name)
            self.tasks[task_name] = task
        else:
            task = self.tasks[task_name]

        if checkpoint_id not in task.checkpoints:
            checkpoint = Checkpoint(
                checkpoint_id=checkpoint_id,
                input_variables={},
                eval_params={},
            )
            task.checkpoints[checkpoint_id] = checkpoint
            return checkpoint
        else:
            return task.checkpoints[checkpoint_id]

    def _save_input_hash(self, task_name: str, checkpoint: Checkpoint) -> None:
        # Remove the checkpoint from any previous input hashes
        task = self.tasks[task_name]
        checkpoint_id = checkpoint.checkpoint_id

        input_hashes_to_remove = []
        for input_hash in task.input_hashes:
            if checkpoint_id in task.input_hashes[input_hash]:
                task.input_hashes[input_hash].remove(checkpoint_id)
            if len(task.input_hashes[input_hash]) == 0:
                input_hashes_to_remove.append(input_hash)

        for input_hash in input_hashes_to_remove:
            del task.input_hashes[input_hash]

        # Add the checkpoint to the new input hash
        input_hash = checkpoint.get_input_hash()
        if input_hash not in task.input_hashes:
            task.input_hashes[input_hash] = set()
        task.input_hashes[input_hash].add(checkpoint_id)

    def record_input_variable(
        self,
        task_name: str,
        checkpoint_id: str,
        variable_name: str,
        value: str,
    ) -> None:
        checkpoint = self._fetch_or_create_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        checkpoint.input_variables[variable_name] = value
        self._save_input_hash(task_name, checkpoint)

    def record_output(self, task_name: str, checkpoint_id: str, output: str) -> None:
        checkpoint = self._fetch_or_create_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        checkpoint.output = output

    def get_latest_checkpoints(
        self, task_name: str, input_hash: str, num_checkpoints: int = 2
    ) -> list[Checkpoint]:
        if task_name not in self.tasks:
            return []
        task = self.tasks[task_name]
        if input_hash not in task.input_hashes:
            return []

        return [
            task.checkpoints[checkpoint_id]
            for checkpoint_id in sorted(task.input_hashes[input_hash])[
                -num_checkpoints:
            ]
        ]

    def get_checkpoint(
        self, task_name: str, checkpoint_id: str
    ) -> Optional[Checkpoint]:
        if task_name not in self.tasks:
            return None
        task = self.tasks[task_name]

        if checkpoint_id not in task.checkpoints:
            return None

        return task.checkpoints[checkpoint_id]

    def get_input_hashes(self, task_name: str) -> list[str]:
        task = self.tasks.get(task_name)
        if task is None:
            return []

        hashes_to_remove = []
        for input_hash in task.input_hashes:
            if len(task.input_hashes[input_hash]) == 0:
                hashes_to_remove.append(input_hash)

        for input_hash in hashes_to_remove:
            del task.input_hashes[input_hash]

        return list(task.input_hashes.keys())

    def record_eval_params(
        self,
        task_name: str,
        checkpoint_id: str,
        eval_params: dict[str, Any],
        rerun_metadata: Optional[dict[str, Any]] = None,
    ) -> None:
        checkpoint = self._fetch_or_create_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        checkpoint.eval_params.update(eval_params)
        if rerun_metadata is not None:
            checkpoint.rerun_metadata = rerun_metadata

    def record_output_feedback(
        self,
        task_name: str,
        checkpoint_id: str,
        feedback: OutputFeedback,
    ) -> None:
        checkpoint = self._fetch_or_create_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        checkpoint.output_feedback = feedback

    def get_task_names(self) -> list[str]:
        return list(self.tasks.keys())

    def record_comparison_result(
        self,
        task_name: str,
        input_hash: str,
        result: ComparisonResult,
    ) -> None:
        task = self.tasks[task_name]
        key = f"{result.older_checkpoint_id},{result.newer_checkpoint_id}"
        task.comparison_results[key] = result

        if input_hash not in task.input_hash_to_comparison_results:
            task.input_hash_to_comparison_results[input_hash] = set()
        task.input_hash_to_comparison_results[input_hash].add(key)

    def record_output_score(
        self,
        task_name: str,
        checkpoint_id: str,
        score: OutputScore,
    ) -> None:
        checkpoint = self._fetch_or_create_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )
        checkpoint.output_score = score

    def get_comparison_result(
        self,
        task_name: str,
        older_checkpoint_id: str,
        newer_checkpoint_id: str,
    ) -> Optional[ComparisonResult]:
        task = self.tasks[task_name]
        key = f"{older_checkpoint_id},{newer_checkpoint_id}"
        if key not in task.comparison_results:
            return None
        return task.comparison_results[key]

    def get_comparison_results_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
        num_results: int = 3,
    ) -> list[ComparisonResult]:
        task = self.tasks[task_name]
        if input_hash not in task.input_hash_to_comparison_results:
            return []
        comparison_results = task.input_hash_to_comparison_results[input_hash]
        return [task.comparison_results[key] for key in comparison_results][
            :num_results
        ]

    def delete_checkpoints_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
    ) -> None:
        task = self.tasks[task_name]
        if input_hash not in task.input_hashes:
            return
        for checkpoint_id in task.input_hashes[input_hash]:
            del task.checkpoints[checkpoint_id]
        del task.input_hashes[input_hash]


class FileRecorder(EvalRecorder):
    def __init__(self, dir_path: str) -> None:
        self.dir_path = dir_path
        if not os.path.exists(self.dir_path):
            os.makedirs(self.dir_path)
        self.yaml_dicts: dict[str, dict[str, str]] = {}

    def _record_checkpoint(
        self,
        task_name: str,
        checkpoint_id: str,
        prefixes: list[str],
        name: str,
        value: Any,
        flush: bool = False,
    ) -> dict[str, Any]:
        dir_name = os.path.join(self.dir_path, task_name, "checkpoints")
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        file_name = os.path.join(dir_name, f"{checkpoint_id}.yaml")

        if checkpoint_id not in self.yaml_dicts:
            if os.path.exists(file_name):
                yaml_dict = yaml.load(open(file_name, "r"), Loader=yaml.FullLoader)
            else:
                yaml_dict = {}
        else:
            yaml_dict = self.yaml_dicts[checkpoint_id]

        current_dict = yaml_dict
        for prefix in prefixes:
            if prefix not in current_dict:
                current_dict[prefix] = {}
            current_dict = current_dict[prefix]
        current_dict[name] = value
        if flush:
            yaml.dump(yaml_dict, open(file_name, "w+"))
            try:
                del self.yaml_dicts[checkpoint_id]
            except KeyError:
                pass
        else:
            self.yaml_dicts[checkpoint_id] = dict(yaml_dict)
        return yaml_dict

    def record_input_variable(
        self, task_name: str, checkpoint_id: str, variable_name: str, value: str
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=["input_variables"],
            name=variable_name,
            value=value,
        )

    def record_output(
        self,
        task_name: str,
        checkpoint_id: str,
        output: str,
    ) -> None:
        yaml_dict = self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="output",
            value=output,
        )

        input_hash = get_input_hash(yaml_dict["input_variables"])
        inputs_dir_name = os.path.join(self.dir_path, task_name, "inputs")
        if not os.path.exists(inputs_dir_name):
            os.makedirs(inputs_dir_name)

        input_file_name = os.path.join(
            self.dir_path, task_name, "inputs", f"{input_hash}.yaml"
        )
        yaml.dump(yaml_dict["input_variables"], open(input_file_name, "w+"))
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="input_hash",
            value=input_hash,
            flush=True,
        )

    def record_eval_params(
        self,
        task_name: str,
        checkpoint_id: str,
        eval_params: dict[str, Any],
        rerun_metadata: Optional[dict[str, Any]] = None,
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="eval_params",
            value=eval_params,
        )
        if rerun_metadata is not None:
            self._record_checkpoint(
                task_name=task_name,
                checkpoint_id=checkpoint_id,
                prefixes=[],
                name="rerun_metadata",
                value=rerun_metadata,
            )

    def record_output_feedback(
        self,
        task_name: str,
        checkpoint_id: str,
        feedback: OutputFeedback,
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="output_feedback",
            value=feedback.as_dict(),
            flush=True,
        )

    def record_output_score(
        self,
        task_name: str,
        checkpoint_id: str,
        score: OutputScore,
    ) -> None:
        self._record_checkpoint(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            prefixes=[],
            name="output_score",
            value=score.as_dict(),
            flush=True,
        )

    def get_checkpoint(
        self, task_name: str, checkpoint_id: str
    ) -> Optional[Checkpoint]:
        file_name = os.path.join(
            self.dir_path, task_name, "checkpoints", f"{checkpoint_id}.yaml"
        )
        if not os.path.exists(file_name):
            return None
        yaml_dict = yaml.load(open(file_name, "r"), Loader=yaml.FullLoader)
        checkpoint = Checkpoint(checkpoint_id=checkpoint_id)
        checkpoint.input_variables = yaml_dict.get("input_variables", {})
        checkpoint.eval_params = yaml_dict.get("eval_params", {})
        checkpoint.output = yaml_dict.get("output")
        checkpoint.output_feedback = (
            OutputFeedback.from_dict(yaml_dict.get("output_feedback"))
            if "output_feedback" in yaml_dict
            else None
        )
        checkpoint.output_score = (
            OutputScore.from_dict(yaml_dict.get("output_score"))
            if "output_score" in yaml_dict
            else None
        )
        checkpoint.rerun_metadata = yaml_dict.get("rerun_metadata")
        return checkpoint

    def _read_checkpoints(self, task_name: str) -> dict[str, list[Checkpoint]]:
        input_hashes_to_checkpoints: dict[str, list[Checkpoint]] = {}
        dir_name = os.path.join(self.dir_path, task_name, "checkpoints")
        for file_name in os.listdir(dir_name):
            if file_name.endswith(".yaml"):
                file_path = os.path.join(dir_name, file_name)
                yaml_dict = yaml.load(open(file_path, "r"), Loader=yaml.FullLoader)
                if input_hash := yaml_dict.get("input_hash"):
                    if input_hash not in input_hashes_to_checkpoints:
                        input_hashes_to_checkpoints[input_hash] = []
                    input_hashes_to_checkpoints[input_hash].append(
                        Checkpoint(**yaml_dict)
                    )

        for input_hash in input_hashes_to_checkpoints:
            input_hashes_to_checkpoints[input_hash].sort(reverse=True)
        return input_hashes_to_checkpoints

    def get_latest_checkpoints(
        self, task_name: str, input_hash: str, num_checkpoints: int = 2
    ) -> list[Checkpoint]:
        input_hashes_to_checkpoints = self._read_checkpoints(task_name)
        if input_hash not in input_hashes_to_checkpoints:
            return []
        return input_hashes_to_checkpoints[input_hash][:num_checkpoints]

    def get_input_hashes(self, task_name: str) -> list[str]:
        input_hashes = []
        for file_name in os.listdir(os.path.join(self.dir_path, task_name, "inputs")):
            if file_name.endswith(".yaml"):
                input_hashes.append(file_name[:-5])
        return input_hashes

    def get_task_names(self) -> list[str]:
        task_names = []
        for dir_name in os.listdir(self.dir_path):
            if os.path.isdir(os.path.join(self.dir_path, dir_name)):
                task_names.append(dir_name)
        return task_names

    def record_comparison_result(
        self,
        task_name: str,
        input_hash: str,
        result: ComparisonResult,
    ) -> None:
        comparison_dir = os.path.join(self.dir_path, task_name, "comparison_results")
        if not os.path.exists(comparison_dir):
            os.makedirs(comparison_dir)

        file_name = f"{input_hash}_{result.older_checkpoint_id}_{result.newer_checkpoint_id}.yaml"
        file_path = os.path.join(comparison_dir, file_name)
        yaml.dump(
            result.output_feedback.as_dict(),
            open(file_path, "w+"),
        )

    def get_comparison_result(
        self,
        task_name: str,
        older_checkpoint_id: str,
        newer_checkpoint_id: str,
    ) -> Optional[ComparisonResult]:
        comparison_dir = os.path.join(self.dir_path, task_name, "comparison_results")
        if not os.path.exists(comparison_dir):
            return None

        for file_name in os.listdir(comparison_dir):
            splits = os.path.splitext(file_name)[0].split("_")
            if len(splits) != 3:
                continue

            if splits[1] != older_checkpoint_id:
                continue

            if splits[2] != newer_checkpoint_id:
                continue

            if file_name.endswith(".yaml"):
                file_path = os.path.join(comparison_dir, file_name)
                yaml_dict = yaml.load(open(file_path, "r"), Loader=yaml.FullLoader)
                return ComparisonResult(
                    older_checkpoint_id=older_checkpoint_id,
                    newer_checkpoint_id=newer_checkpoint_id,
                    output_feedback=OutputFeedback.from_dict(yaml_dict),
                )
        return None

    def get_comparison_results_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
        num_results: int = 3,
    ) -> list[ComparisonResult]:
        comparison_dir = os.path.join(self.dir_path, task_name, "comparison_results")
        if not os.path.exists(comparison_dir):
            return []

        results: list[ComparisonResult] = []
        for file_name in os.listdir(comparison_dir):
            splits = os.path.splitext(file_name)[0].split("_")
            if len(splits) != 3:
                continue

            if splits[0] != input_hash:
                continue

            if file_name.endswith(".yaml"):
                file_path = os.path.join(comparison_dir, file_name)
                yaml_dict = yaml.load(open(file_path, "r"), Loader=yaml.FullLoader)
                results.append(
                    ComparisonResult(
                        older_checkpoint_id=splits[1],
                        newer_checkpoint_id=splits[2],
                        output_feedback=OutputFeedback.from_dict(yaml_dict),
                    )
                )
        return sorted(results, key=lambda x: x.newer_checkpoint_id, reverse=True)[
            :num_results
        ]

    def delete_checkpoints_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
    ) -> None:
        ...


class DiskRecorder(EvalRecorder):
    # TODO: improve this by using rocksdb etc. vs a stupid pickle file
    def __init__(self, dir_path: str) -> None:
        self.file_name = os.path.join(dir_path, "eyeball_checkpoints.pkl")
        if os.path.exists(self.file_name):
            self.memory_recorder = pickle.load(open(self.file_name, "rb"))
        else:
            self.memory_recorder = MemoryRecorder()

    def record_input_variable(
        self,
        task_name: str,
        checkpoint_id: str,
        variable_name: str,
        value: str,
    ) -> None:
        self.memory_recorder.record_input_variable(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            variable_name=variable_name,
            value=value,
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def record_eval_params(
        self,
        task_name: str,
        checkpoint_id: str,
        eval_params: dict[str, Any],
        rerun_metadata: Optional[dict[str, Any]] = None,
    ) -> None:
        self.memory_recorder.record_eval_params(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            eval_params=eval_params,
            rerun_metadata=rerun_metadata,
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def record_output_feedback(
        self,
        task_name: str,
        checkpoint_id: str,
        feedback: OutputFeedback,
    ) -> None:
        self.memory_recorder.record_output_feedback(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            feedback=feedback,
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def get_checkpoint(
        self, task_name: str, checkpoint_id: str
    ) -> Optional[Checkpoint]:
        return self.memory_recorder.get_checkpoint(
            task_name=task_name, checkpoint_id=checkpoint_id
        )

    def get_latest_checkpoints(
        self, task_name: str, input_hash: str, num_checkpoints: int = 2
    ) -> list[Checkpoint]:
        return self.memory_recorder.get_latest_checkpoints(
            task_name=task_name, input_hash=input_hash, num_checkpoints=num_checkpoints
        )

    def get_input_hashes(self, task_name: str) -> list[str]:
        return self.memory_recorder.get_input_hashes(task_name)

    def get_task_names(self) -> list[str]:
        return self.memory_recorder.get_task_names()

    def record_comparison_result(
        self,
        task_name: str,
        input_hash: str,
        result: ComparisonResult,
    ) -> None:
        self.memory_recorder.record_comparison_result(
            task_name=task_name, input_hash=input_hash, result=result
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def record_output_score(
        self,
        task_name: str,
        checkpoint_id: str,
        score: OutputScore,
    ) -> None:
        self.memory_recorder.record_output_score(
            task_name=task_name, checkpoint_id=checkpoint_id, score=score
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def get_comparison_result(
        self,
        task_name: str,
        older_checkpoint_id: str,
        newer_checkpoint_id: str,
    ) -> Optional[ComparisonResult]:
        return self.memory_recorder.get_comparison_result(
            task_name=task_name,
            older_checkpoint_id=older_checkpoint_id,
            newer_checkpoint_id=newer_checkpoint_id,
        )

    def get_comparison_results_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
        num_results: int = 3,
    ) -> list[ComparisonResult]:
        return self.memory_recorder.get_comparison_results_for_input_hash(
            task_name=task_name, input_hash=input_hash, num_results=num_results
        )

    def record_output(
        self,
        task_name: str,
        checkpoint_id: str,
        output: str,
    ) -> None:
        self.memory_recorder.record_output(
            task_name=task_name,
            checkpoint_id=checkpoint_id,
            output=output,
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def delete_checkpoints_for_input_hash(
        self,
        task_name: str,
        input_hash: str,
    ) -> None:
        self.memory_recorder.delete_checkpoints_for_input_hash(
            task_name=task_name, input_hash=input_hash
        )
        pickle.dump(self.memory_recorder, open(self.file_name, "wb"))

    def compute_latest_comparison_results(self, task_name: str) -> None:
        ...



from .classes import FeedbackResult, OutputFeedback, OutputScore
import openai


def output_feedback_from_scores(
    older_score: OutputScore, newer_score: OutputScore
) -> OutputFeedback:
    if older_score == newer_score:
        return OutputFeedback(
            FeedbackResult.NEUTRAL,
            f"Score {older_score} is equal to {newer_score}",
        )
    elif older_score > newer_score:
        return OutputFeedback(
            FeedbackResult.NEGATIVE,
            f"Score {newer_score} is worse than {newer_score}",
        )
    else:
        return OutputFeedback(
            FeedbackResult.POSITIVE,
            f"Score {newer_score} is better than {older_score}",
        )


def model_graded_comparator(
    objective: str,
    input_variables: dict[str, str],
    older_checkpoint_output: str,
    newer_checkpoint_output: str,
) -> OutputFeedback:
    input_str = "\n".join([f"{key}: {val}" for key, val in input_variables.items()])

    system_msg = f"""
You are a human evaluator trying to grade the response of a function based on the following objective and inputs.

Objective:
{objective}

Inputs: 
{input_str}
"""

    is_b_better = f"""
Keeping the objectives and the inputs in mind, which of the following responses is better?

Response A:
{older_checkpoint_output}

Response B:
{newer_checkpoint_output}

Give your reasoning followed by one of the following options:
Yes -- if Response A is better than B
No -- if Response B is better than A
Same -- if both are equally good
"""
    response = openai.ChatCompletion.create(  # type: ignore
        model="gpt-4",
        temperature=0.5,
        messages=[
            {"role": "system", "content": system_msg},
            {"role": "user", "content": is_b_better},
        ],
    )["choices"][0]["message"]["content"]
    response_last_line = response.split("\n")[-1].lower()
    if "yes" in response_last_line:
        return OutputFeedback(FeedbackResult.NEGATIVE, response)
    elif "no" in response_last_line:
        return OutputFeedback(FeedbackResult.POSITIVE, response)
    else:
        return OutputFeedback(FeedbackResult.NEUTRAL, response)



from enum import IntEnum
from typing import Any, Callable
from dataclasses import dataclass
from typing import Protocol


class FeedbackResult(IntEnum):
    POSITIVE = 1
    NEGATIVE = -1
    NEUTRAL = 0

    def __str__(self) -> str:
        if self == FeedbackResult.POSITIVE:
            return "📈 +ve"
        elif self == FeedbackResult.NEGATIVE:
            return "📉 -ve"
        else:
            return "📊 neutral"


@dataclass
class OutputFeedback:
    result: FeedbackResult
    message: str

    def as_dict(self):
        return {"result": self.result.name, "message": self.message}

    @staticmethod
    def from_dict(data):
        return OutputFeedback(
            result=FeedbackResult[data["result"]], message=data["message"]
        )


@dataclass
class OutputScore:
    score: float
    message: str

    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, OutputScore) and self.score == __value.score

    def __lt__(self, __value: object) -> bool:
        return isinstance(__value, OutputScore) and self.score < __value.score

    def __gt__(self, __value: object) -> bool:
        return isinstance(__value, OutputScore) and self.score > __value.score

    def __str__(self) -> str:
        return f"{self.score:.2f} ({self.message})"

    def as_dict(self):
        return {"score": self.score, "message": self.message}

    @staticmethod
    def from_dict(data):
        return OutputScore(score=data["score"], message=data["message"])


# Compare the output of two checkpoints for a given objective
# The comparator should return the output of -- How would you rate the newer_checkpoint output compared to older_checkpoint output? Is it positive, negative or neutral?
class OutputComparator(Protocol):
    def __call__(
        self,
        objective: str,
        input_variables: dict[str, str],
        older_checkpoint_output: str,
        newer_checkpoint_output: str,
    ) -> OutputFeedback:
        ...


# The output scorer is used to score the output of a model given the objective and inputs
# The scorer should return a float with a higher value indicating a better output
class OutputScorer(Protocol):
    def __call__(
        self, objective: str, input_variables: dict[str, str], output: str
    ) -> OutputScore:
        ...



from typing import Optional
from fire import Fire


def compare_checkpoints(
    task_name: Optional[str] = None,
    objective: Optional[str] = None,
):
    pass


if __name__ == "__main__":
    pass
